#ifdef BEGIN_MAPPER
#include "mappers.hh" // SECOND INCLUSION

map<unsigned, mapper *>mappers;

BEGIN_MAPPER(0, )
	virtual void write(fWORD ofs, fBYTE value)
	{
        fprintf(stderr, "Unknown mapper wrote %02X to %04ZX\n", value, ofs);
    }
END_MAPPER(0)

BEGIN_MAPPER(1, pos=reg=0;memset(regs,0,sizeof regs);regs[0]=0xC)
	BYTE reg;
	BYTE pos;
	BYTE regs[4];
	
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 1: [%04X] <- %02X\n", ofs, value);
		if(value&0x80)
			pos=reg=0;
		else
		{
			reg |= (value&1) << pos;
			if(++pos==5)
			{
				if(debug)fprintf(stderr, "Mapper 1: Reg value therefore %02X\n", reg);
				regs[ofs / 0x2000 - 4] = reg;
				rewrite();
				pos = reg = 0;
			}
		}
	}
	void rewrite()
	{
		reset();
		if(regs[0] & 8)
			maprom(4, (regs[0] & 4) ? 0 : 4, regs[3]);
		else
			maprom(8, 0, regs[3]);
		update();
	}
END_MAPPER(1)

BEGIN_MAPPER(2, )
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 2: [%04X] <- %02X\n", ofs, value);
		maprom(4, 0, value);
		update();
	}
END_MAPPER(2)

BEGIN_MAPPER(4, )
	BYTE cmd;
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 4: [%04X] <- %02X\n", ofs, value);
		if(ofs==0x8000)
		{
			exchange(0, 2, 0, 4, value&64); // Swap $8000 and $C000 (8k)
			
			cmd = value & 7;
			update();
		}
		else if(ofs==0x8001)
		{
			switch(cmd)
			{
				case 6: maprom(2, 0, value); break; // Map 8k at $8000
				case 7: maprom(2, 2, value); break; // Map 8k at $A000
			}
			update();
		}
	}
END_MAPPER(4)

BEGIN_MAPPER(7, )
	virtual void write(fWORD, fBYTE value)
	{
		maprom(8, 0, value & 15);
		update();
	}
END_MAPPER(7)

BEGIN_MAPPER(15, debug=true; maprom(8, 0, 0))
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 15: [%04X] <- %02X\n", ofs, value);
		switch(ofs)
		{
			case 0x8000:
			{
				unsigned i = value&63;
				maprom(4, 0, i);
				maprom(4, 4, i+1 );
				exchange(0, 2, 0, 2, (value&128));
				exchange(1, 2, 4, 6, !(value&128));
				break;
			}
			case 0x8001:
				maprom(4, 4, value&63);
				exchange(1, 2, 4, 6, (value&128));
				break;
			case 0x8002:
			{
				unsigned i = value*2;
				reset();
				if(value >> 7)
				{
					++i;
					maprom(2, 2, i);
					maprom(2, 6, i);
				}
				else
				{
					maprom(2, 0, i);
					maprom(2, 4, i);
				}				
				break;
			}
			case 0x8003:
				/* &64 = mirroring control */
				maprom(4, 4, value&63);
				exchange(1, 2, 4, 6, (value&128));
				break;
		}
		update();
	}
END_MAPPER(15)

#else
#error mappers.cc is not to be compiled individually!
#endif
